@@ -13,7 +13,12 @@ import copy
 import orjson
 from hashlib import blake2b
 from tethys_utils.data_models import Geometry, Dataset, DatasetBase, Station, Stats, StationBase
+# from data_models import Geometry, Dataset, DatasetBase, Station, Stats, StationBase
 from tethys_utils.misc import make_run_date_key, grp_ts_agg, write_pkl_zstd
+# from misc import make_run_date_key, grp_ts_agg, write_pkl_zstd
+import geojson
+from shapely.geometry import shape, mapping
+from shapely import wkb
 
 ############################################
 ### Parameters
@@ -26,7 +31,7 @@ agg_stat_mapping = {'mean': 'mean', 'cumulative': 'sum', 'continuous': None, 'ma
 ### Functions
 
 
-def results_data_integrety_checks(data, param_name, attrs, encoding, ancillary_variables=None):
+def results_data_integrety_checks(data, param_name, attrs, encoding, spatial_distribution, ancillary_variables=None):
     """
 
     """
@@ -42,9 +47,15 @@ def results_data_integrety_checks(data, param_name, attrs, encoding, ancillary_v
 
     data_cols.extend([param_name])
 
-    ts_index_list = ['time', 'height']
+    if spatial_distribution == 'sparse':
+        ts_index_list = ['geometry', 'time', 'height']
+    elif spatial_distribution == 'grid':
+        ts_index_list = ['lon', 'lat', 'time', 'height']
+    else:
+        raise ValueError('spatial_distribution should be either sparse or grid.')
+
     ts_essential_list = [param_name]
-    ts_no_attrs_list = ['modified_date']
+    ts_no_attrs_list = ['modified_date', 'lat', 'lon', 'geometry']
 
     ts_data_cols = list(data.columns)
 
@@ -81,9 +92,18 @@ def station_data_integrety_checks(data, attrs=None, encoding=None):
     """
 
     """
-    # Station data
-    ignore_keys = ['station_id', 'lat', 'lon', 'name', 'altitude', 'ref', 'virtual_station', 'geometry']
-    other_keys = [s for s in data if s not in ignore_keys]
+    stn_m = StationBase(**data)
+    stn_data = orjson.loads(stn_m.json(exclude_none=True))
+
+    essential_keys = ['geometry', 'station_id']
+
+    for c in essential_keys:
+        if not c in stn_data:
+            raise ValueError('The Dict must contain the key: ' + str(c))
+
+
+    ignore_attr_keys = ['station_id', 'lat', 'lon', 'name', 'altitude', 'ref', 'virtual_station', 'geometry']
+    other_keys = [s for s in stn_data if s not in ignore_attr_keys]
 
     if other_keys:
 
@@ -101,24 +121,77 @@ def station_data_integrety_checks(data, attrs=None, encoding=None):
         else:
             raise TypeError('encoding must be a dict')
 
-    return data
+    return stn_data
 
 
-def data_to_xarray(results_data, station_data, param_name, results_attrs, results_encoding, station_attrs=None, station_encoding=None, virtual_station=False, run_date=None, ancillary_variables=None, compression=False, compress_level=1):
+def metadata_integrety_checks(results_data, station_data, ds_metadata):
+    """
+
+    """
+    sd = ds_metadata['spatial_distribution']
+    geo_type = ds_metadata['geometry_type']
+    grouping = ds_metadata['grouping']
+
+    data_index = list(results_data.index.names)
+    if ('geometry' in data_index) and (sd == 'grid'):
+        raise ValueError('You have passed geometry as an index, but the spatial_distribution is labeled as grid...something is inconsistent.')
+    elif (sd == 'sparse') and (('lon' in data_index) or ('lat' in data_index)):
+        raise ValueError('You have passed lon/lat as an index, but the spatial_distribution is labeled as sparse...something is inconsistent.')
+
+    stn_geometry = station_data['geometry']
+    stn_geo_type = stn_geometry['type']
+
+    if 'geometry' in data_index:
+        data_sd = 'sparse'
+        hex_geo1 = results_data.index.get_level_values('geometry').unique()
+        if len(hex_geo1) == 1:
+            data_grouping = 'none'
+        else:
+            data_grouping = 'blocks'
+
+        data_geometry = wkb.loads(hex_geo1[0], True)
+        data_geo_type = data_geometry.type
+    elif ('lon' in data_index) and ('lat' in data_index):
+        data_sd = 'grid'
+        data_geo_type = 'Point'
+        lons = results_data.index.get_level_values('lon').unique()
+        lats = results_data.index.get_level_values('lat').unique()
+        if (len(lons) == 1) and (len(lats) == 1):
+            data_grouping = 'none'
+        else:
+            data_grouping = 'blocks'
+    else:
+        raise ValueError('Data is not indexed correctly.')
+
+    if data_grouping == 'blocks':
+        if stn_geo_type != 'Polygon':
+            raise ValueError('If the data grouping is "blocks", then the station geometry should be a Polygon rectangle of the extent.')
+    else:
+        if stn_geo_type != data_geo_type:
+            raise ValueError('If the data grouping is "none", then the station geometry should be the same as the data geometry.')
+
+    if data_geo_type != geo_type:
+        raise ValueError('The data geometry type does not match the geometry type listed in the dataset metadata.')
+    if data_grouping != grouping:
+        raise ValueError('The data grouping does not match the grouping listed in the dataset metadata.')
+    if data_sd != sd:
+        raise ValueError('The data spatial distribution does not match the spatial distribution listed in the dataset metadata.')
+
+
+def data_to_xarray(results_data, station_data, param_name, results_attrs, results_encoding, station_attrs=None, station_encoding=None, run_date=None, ancillary_variables=None, compression=False, compress_level=1):
     """
     Converts DataFrames of time series data, station data, and other attributes to an Xarray Dataset. Optionally has Zstandard compression.
 
     Parameters
     ----------
     results_data : DataFrame
-        DataFrame of the core parameter and associated ancillary variable indexed by time and height.
-        The index should have the names of "time" and "height". "height" is height above the surface. So if the parameter represents a surface measurement, then the height should be 0.
+        DataFrame of the core parameter and associated ancillary variable. If the spatial distribution is sparse then the data should be indexed by geometry, time, and height. If the spatial distribution is grid then the data should be indexed by lon, lat, time, and height.
     station_data : dict
-        Dictionary of the station data. Should include a station_id which should be a hashed string from blake2b (digest_size=12) of the geojson geometry. The minimum necessary other fields should include lat, lon, and altitude. Data owner specific other fields can include "ref" for the reference id and "name" for the station name.
+        Dictionary of the station data. Should include a station_id which should be a hashed string from blake2b (digest_size=12) of the geojson geometry. The other necessary field is geometry, which is the geometry of the results object is not grouped or a boundary extent as a polygon if the results object is grouped. Data owner specific other fields can include "ref" for the reference id and "name" for the station name.
     param_name : str
-        The core parameter name of the column in the ts_data DataFrame.
+        The core parameter name of the column in the results_data DataFrame.
     results_attrs : dict
-        A dictionary of the xarray/netcdf attributes of the results_data. Where the keys are the columns and the values are the attributes.
+        A dictionary of the xarray/netcdf attributes of the results_data. Where the keys are the columns and the values are the attributes. Only necessary if additional ancilliary valiables are added to the results_data.
     results_encoding : dict
         A dictionary of the xarray/netcdf encodings for the results_data.
     station_attrs : dict or None
@@ -129,30 +202,40 @@ def data_to_xarray(results_data, station_data, param_name, results_attrs, result
     Returns
     -------
     Xarray Dataset or bytes object
-
     """
-    ## Station data prep
-    stn_m = StationBase(**station_data)
-    stn_data = orjson.loads(stn_m.json(exclude_none=True))
-    stn_data['lon'] = stn_data['geometry']['coordinates'][0]
-    stn_data['lat'] = stn_data['geometry']['coordinates'][1]
+    ## dataset metadata
+    ds_mapping = results_attrs[param_name]
+    param_name = ds_mapping['parameter']
+    sd = ds_mapping['spatial_distribution']
+    grouping = ds_mapping['grouping']
+    geo_type = ds_mapping['geometry_type']
+
+    ## Integrity Checks
+
+    # Time series data
+    ts_data1 = results_data_integrety_checks(results_data, param_name, results_attrs, results_encoding, sd, ancillary_variables)
+
+    # Station data
+    stn_data = station_data_integrety_checks(station_data, attrs=station_attrs, encoding=station_encoding)
+
+    # Metadata
+    metadata_integrety_checks(ts_data1, stn_data, ds_mapping)
+
+    ## geometry from station data
+    stn_geometry = stn_data['geometry']
     stn_data.pop('geometry')
+    if grouping == 'blocks':
+        stn_data['extent'] = shape(stn_geometry).wkb_hex
+
+    ## Station properties
     if 'properties' in stn_data:
         props = stn_data['properties'].copy()
         for k, v in props.items():
             stn_data[k] = v
         stn_data.pop('properties')
 
-    ## Integrity Checks
-
-    # Time series data
-    ts_data1 = results_data_integrety_checks(results_data, param_name, results_attrs, results_encoding, ancillary_variables)
-
-    # Station data
-    stn_data = station_data_integrety_checks(stn_data, attrs=station_attrs, encoding=station_encoding)
-
     ## Assign Attributes
-    attrs1 = {'station_id': {'cf_role': "timeseries_id", 'virtual_station': virtual_station}, 'virtual_station': {'long_name': 'Is this station a virtual or modeled station?'}, 'lat': {'standard_name': "latitude", 'units': "degrees_north"}, 'lon': {'standard_name': "longitude", 'units': "degrees_east"}, 'altitude': {'standard_name': 'surface_altitude', 'long_name': 'height above the geoid to the lower boundary of the atmosphere', 'units': 'm'}}
+    attrs1 = {'station_id': {'cf_role': "timeseries_id"}, 'lat': {'standard_name': "latitude", 'units': "degrees_north"}, 'lon': {'standard_name': "longitude", 'units': "degrees_east"}, 'altitude': {'standard_name': 'surface_altitude', 'long_name': 'height above the geoid to the lower boundary of the atmosphere', 'units': 'm'}, 'geometry': {'long_name': 'The hexadecimal encoding of the Well-Known Binary (WKB) geometry', 'crs_EPSG': 4326}, 'extent': {'long_name': 'The hexadecimal encoding of the Well-Known Binary (WKB) station extent', 'crs_EPSG': 4326}}
 
     if 'name' in stn_data.keys():
         attrs1.update({'name': {'long_name': 'station name'}})
@@ -162,7 +245,7 @@ def data_to_xarray(results_data, station_data, param_name, results_attrs, result
     if isinstance(station_attrs, dict):
         attrs1.update(copy.deepcopy(station_attrs))
 
-    ts_cols = list(results_data.columns)
+    ts_cols = list(ts_data1.columns)
 
     attrs1.update(results_attrs)
     if 'cf_standard_name' in attrs1[param_name]:
@@ -181,11 +264,10 @@ def data_to_xarray(results_data, station_data, param_name, results_attrs, result
     if isinstance(station_encoding, dict):
         encoding1.update(copy.deepcopy(station_encoding))
 
-    height = pd.to_numeric(results_data.reset_index()['height'], downcast='integer')
+    height = pd.to_numeric(ts_data1.reset_index()['height'], downcast='integer')
 
     if 'int' in height.dtype.name:
-        height_enc = {'dtype': height.dtype.name, '_FillValue': -9999}
-        # dtype = height.dtype.name
+        height_enc = {'dtype': height.dtype.name}
     elif 'float' in height.dtype.name:
         height_enc = {'dtype': 'int32', '_FillValue': -9999, 'scale_factor': 0.001}
     else:
@@ -197,11 +279,32 @@ def data_to_xarray(results_data, station_data, param_name, results_attrs, result
         encoding1.update({'modified_date': {'_FillValue': -99999999, 'units': "days since 1970-01-01 00:00:00"}})
 
     ## Create the Xarray Dataset
-    ds1 = results_data.to_xarray()
+    data_index = list(ts_data1.index.names)
+    ds1 = ts_data1.to_xarray()
 
     ## Assign the stn data to the main Dataset
+    if 'extent' in stn_data:
+        ds1 = ds1.assign_coords({'extent': [stn_data['extent']]})
+        stn_data.pop('extent')
+
+    if 'extent' in ds1:
+        stn_index = ('extent')
+    else:
+        if 'geometry' in data_index:
+            stn_index = ('geometry')
+        else:
+            stn_index = ('lon', 'lat')
+
+    # ds1 = ds1.assign_coords({'station_id': (stn_index, [stn_data['station_id']])})
+    # stn_data.pop('station_id')
+
     for k, v in stn_data.items():
-        ds1[k] = v
+        ds1 = ds1.assign({k: (stn_index, [v])})
+
+    ## Assign the lat and lon if the data are sparse points
+    if ('geometry' in data_index) and (stn_geometry['type'] == 'Point'):
+        ds1 = ds1.assign({'lon': (stn_index, [stn_geometry['coordinates'][0]])})
+        ds1 = ds1.assign({'lat': (stn_index, [stn_geometry['coordinates'][1]])})
 
     ## Add attributes and encodings
     for e, val in encoding1.items():
@@ -218,11 +321,10 @@ def data_to_xarray(results_data, station_data, param_name, results_attrs, result
         if a in ds1:
             ds1[a].attrs = val
 
-    ds_mapping = results_attrs[param_name]
     title_str = '{agg_stat} {parameter} in {units} of the {feature} by a {method} owned by {owner}'.format(agg_stat=ds_mapping['aggregation_statistic'], parameter=ds_mapping['parameter'], units=ds_mapping['units'], feature=ds_mapping['feature'], method=ds_mapping['method'], owner=ds_mapping['owner'])
 
     run_date_key = make_run_date_key(run_date)
-    ds1.attrs = {'featureType': 'timeSeries', 'title': title_str, 'institution': ds_mapping['owner'], 'license': ds_mapping['license'], 'source': ds_mapping['method'], 'history': run_date_key + ': Generated'}
+    ds1.attrs = {'featureType': 'timeSeries', 'title': title_str, 'institution': ds_mapping['owner'], 'license': ds_mapping['license'], 'source': ds_mapping['method'], 'history': run_date_key + ': Generated', 'version': 3}
 
     ## Test conversion to netcdf
     p_ts1 = ds1.to_netcdf()
@@ -321,14 +423,6 @@ def compare_xrs(old_xr, new_xr, add_old=False):
     parameter = [v for v in vars1 if 'dataset_id' in new_xr[v].attrs][0]
     vars2 = [parameter]
 
-    if not parameter in old_xr:
-        raise ValueError(parameter + ' must be in old_xr')
-
-    on = new_xr[parameter].dims
-
-    if not on == old_xr[parameter].dims:
-        raise ValueError('Dimensions are not the same between the datasets')
-
     ## Determine if there are ancillary variables to pull through
     new_attrs = new_xr[parameter].attrs.copy()
 
@@ -336,11 +430,39 @@ def compare_xrs(old_xr, new_xr, add_old=False):
         av1 = new_attrs['ancillary_variables'].split(' ')
         vars2.extend(av1)
 
-    ## Pull out data for comparison
-    old_df = old_xr[vars2].to_dataframe().reset_index()
-    new_df = new_xr[vars2].to_dataframe().reset_index()
+    if not parameter in old_xr:
+        raise ValueError(parameter + ' must be in old_xr')
 
-    # old_df['modified_date'] = pd.Timestamp('2020-12-29')
+    ## Reduce the dimensions for the comparison for compatibility
+    new1_s = new_xr[vars2].squeeze()
+    old1_s = old_xr[vars2].squeeze()
+    on = list(new1_s.dims)
+
+    if not on == list(old1_s.dims):
+        raise ValueError('Dimensions are not the same between the datasets')
+
+    keep_vars = on + vars2
+
+    new_all_vars = list(new1_s.variables)
+    new_bad_vars = [v for v in new_all_vars if not v in keep_vars]
+    new2_s = new1_s.drop(new_bad_vars)
+
+    old_all_vars = list(old1_s.variables)
+    old_bad_vars = [v for v in old_all_vars if not v in keep_vars]
+    old2_s = old1_s.drop(old_bad_vars)
+
+    # Fix datetime rounding issues...
+    for v in list(old2_s.variables):
+        if old2_s[v].dtype.name == 'datetime64[ns]':
+            old2_s[v] = old2_s[v].dt.round('s')
+
+    for v in list(new2_s.variables):
+        if new2_s[v].dtype.name == 'datetime64[ns]':
+            new2_s[v] = new2_s[v].dt.round('s')
+
+    ## Pull out data for comparison
+    old_df = old2_s.to_dataframe().reset_index()
+    new_df = new2_s.to_dataframe().reset_index()
 
     ## run comparison
     comp = compare_dfs(old_df, new_df, on, parameter, add_old=add_old)
@@ -354,14 +476,28 @@ def compare_xrs(old_xr, new_xr, add_old=False):
         ## Repackage into netcdf
         comp2 = comp.set_index(list(on)).sort_index().to_xarray()
 
+        # Fix datetime rounding issues...
+        for v in list(comp2.variables):
+            if comp2[v].dtype.name == 'datetime64[ns]':
+                comp2[v] = comp2[v].dt.round('s')
+
         for v in vars1:
-            if v in comp2:
+            if v not in vars2:
+                comp2[v] = new_xr[v].copy()
                 comp2[v].attrs = new_xr[v].attrs.copy()
                 comp2[v].encoding = new_xr[v].encoding.copy()
-            else:
-                comp2[v] = new_xr[v].copy()
+
+        new_dims = new_xr[parameter].dims
+        dim_dict = dict(comp2.dims)
+        data_shape = tuple(dim_dict[d] for d in new_dims)
+
+        for v in vars2:
+            comp2 = comp2.assign({v: (new_dims, comp2[v].values.reshape(data_shape))})
+            comp2[v].attrs = new_xr[v].attrs.copy()
+            comp2[v].encoding = new_xr[v].encoding.copy()
 
         comp2.attrs = new_xr.attrs.copy()
+        comp2.encoding = new_xr.encoding.copy()
 
         return comp2
 
@@ -406,24 +542,35 @@ def process_datasets(datasets):
     return dataset_list
 
 
-def create_geometry(coords, geo_type='Point'):
+def create_geometry_df(df, extent=False, altitude=True, precision=7):
     """
 
     """
-    # print(type(coords[0]))
-    if geo_type == 'Point':
-        coords = [np.round(coords[0], 5), np.round(coords[1], 5)]
-        # geo1 = Point(coords)
-        geo1 = {"coordinates": coords, "type": "Point"}
+    if extent:
+        if ('lon' in df) and ('lon' in df):
+            min_lon = df['lon'].min()
+            max_lon = df['lon'].max()
+            min_lat = df['lat'].min()
+            max_lat = df['lat'].max()
+            geometry = geojson.Polygon([[(min_lon, min_lat), (min_lon, max_lat), (max_lon, max_lat), (max_lon, min_lat), (min_lon, min_lat)]], True, precision=precision)
+        else:
+            raise ValueError('Extent must have lat and lon in the df.')
     else:
-        raise ValueError('geo_type not implemented yet')
+        if 'geometry' in df:
+            geometry = df['geometry']
+        elif ('lon' in df) and ('lon' in df):
+            if altitude:
+                if 'altitude' in df:
+                    coords = df.apply(lambda x: (x.lon, x.lat, x.altitude), axis=1)
+                else:
+                    coords = df.apply(lambda x: (x.lon, x.lat), axis=1)
+            else:
+                coords = df.apply(lambda x: (x.lon, x.lat), axis=1)
+            geometry = coords.apply(lambda x: geojson.Point(x, True, precision=precision))
+        else:
+            raise ValueError('Either a dict of geometry or a combo of lat and lon must be in the dataframe.')
 
-    # if not geo1.is_valid:
-    #     raise ValueError('coordinates are not valid')
-
-    geo2 = Geometry(**geo1).dict()
-
-    return geo2
+    return geometry
 
 
 def assign_station_id(geometry):
@@ -435,6 +582,17 @@ def assign_station_id(geometry):
     return station_id
 
 
+def assign_station_ids_df(stns_df, extent=False, precision=5):
+    """
+
+    """
+    geometry = create_geometry_df(stns_df, extent, False, precision)
+
+    stn_ids = geometry.apply(lambda x: assign_station_id(x))
+
+    return stn_ids
+
+
 def get_new_stats(data):
     """
 
@@ -457,34 +615,34 @@ def get_new_stats(data):
     return stats1
 
 
-def process_stations_df(stns_df):
-    """
-    For processing the initial stations prior to assigning the altitude.
-    """
-    stns1 = stns_df.copy()
-    stns1['coords'] = stns1.apply(lambda x: [x.lon, x.lat], axis=1)
+# def process_stations_df(stns_df, precision=6):
+#     """
+#     For processing the initial stations prior to assigning the altitude.
+#     """
+#     stns1 = stns_df.copy()
+#     stns1['coords'] = stns1.apply(lambda x: [x.lon, x.lat], axis=1)
 
-    stns1['geo'] = stns1.apply(lambda x: create_geometry(x.coords), axis=1)
-    stns1['station_id'] = stns1.apply(lambda x: assign_station_id(x.geo), axis=1)
+#     stns1['geometry'] = stns1.apply(lambda x: geojson.Point(x.coords, True, precision=precision), axis=1)
+#     stns1['station_id'] = stns1.apply(lambda x: assign_station_id(x.geometry), axis=1)
 
-    stns2 = stns1.drop_duplicates('station_id').drop('geo', axis=1).copy()
+#     stns2 = stns1.drop_duplicates('station_id').drop(['coords', 'lat', 'lon'], axis=1).copy()
 
-    return stns2
+#     return stns2
 
 
-def process_station_base(coords=None, station_id=None, geometry=None, ref=None, name=None, osm_id=None, altitude=None, properties=None, virtual_station=False, geo_type='Point', return_dict=True):
+def process_station_base(coords=None, station_id=None, geometry=None, ref=None, name=None, osm_id=None, altitude=None, properties=None, return_dict=True):
     """
 
     """
     if isinstance(coords, list):
         ## Create geometry and station_id
-        geometry = create_geometry(coords, geo_type=geo_type)
+        geometry = geojson.Point(coords, True, precision=5)
         station_id = assign_station_id(geometry)
     elif (not isinstance(station_id, str)) & (not isinstance(geometry, dict)):
         raise ValueError('coords or station_id must be assigned')
 
     ## Put into data model
-    stn_m = StationBase(station_id=station_id, geometry=geometry, ref=ref, name=name, osm_id=osm_id, altitude=altitude, properties=properties, virtual_station=virtual_station)
+    stn_m = StationBase(station_id=station_id, geometry=geometry, ref=ref, name=name, osm_id=osm_id, altitude=altitude, properties=properties)
 
     if return_dict:
         stn_m = orjson.loads(stn_m.json(exclude_none=True))
@@ -520,24 +678,34 @@ def get_station_data_from_xr(data):
         data_vars.extend(ancillary_variables)
 
     stn_fields = list(StationBase.schema()['properties'].keys())
-    stn_fields.extend(['lon', 'lat'])
 
-    stn_vars = [v for v in vars1 if (not v in dims1) and (not v in data_vars)]
-    stn_data1 = {k: v['data'] for k, v in data[stn_vars].to_dict()['data_vars'].items() if k in stn_fields}
-    stn_data1['geometry'] = create_geometry([stn_data1['lon'], stn_data1['lat']])
-    stn_data1.pop('lon')
-    stn_data1.pop('lat')
-    stn_data1['altitude'] = round(stn_data1['altitude'], 3)
-    if not 'virtual_station' in stn_data1:
-        stn_data1['virtual_station'] = False
+    ## Geometry
+    if 'geometry' in dims1:
+        if len(data['geometry']) == 1:
+            geo1 = mapping(wkb.loads(data['geometry'].values[0], True))
+        else:
+            geo1 = mapping(wkb.loads(data['extent'].values[0], True))
+
+        stn_fields.remove('geometry')
+
+    lat_lon = ['lon', 'lat']
+
+    stn_vars = [v for v in vars1 if (not v in dims1) and (not v in data_vars) and (not v in lat_lon)]
+    stn_data1 = {k: v['data'][0] for k, v in data[stn_vars].to_dict()['data_vars'].items() if k in stn_fields}
+    stn_data1.update({'station_id': data['station_id'].values[0], 'geometry': geo1})
+    if 'altitude' in stn_data1:
+        stn_data1['altitude'] = round(stn_data1['altitude'], 3)
+    # if not 'virtual_station' in stn_data1:
+    #     stn_data1['virtual_station'] = False
 
     props = {s: {'data': data[s].to_dict()['data'], 'attrs': data[s].to_dict()['attrs']} for s in stn_vars if s not in stn_fields}
-    stn_data1['properties'] = props
+    if props:
+        stn_data1['properties'] = props
 
     ## Check model
     stn_m = StationBase(**stn_data1)
 
-    return stn_data1
+    return stn_m.dict(exclude_none=True)
 
 
 def process_station_summ(dataset_id, station_id, data, object_infos, mod_date=None):
@@ -559,10 +727,12 @@ def process_station_summ(dataset_id, station_id, data, object_infos, mod_date=No
 
     station_m = Station(**stn_dict2)
 
-    return station_m
+    stn_dict = orjson.loads(station_m.json(exclude_none=True))
+
+    return stn_dict
 
 
-def prepare_results(data_dict, dataset_list, station_dict, data_df, run_date_key, mod_date=None, sum_closed='right', other_closed='left', discrete=True, station_attrs=None, station_encoding=None):
+def prepare_results(data_dict, dataset_list, station_data, results_data, run_date_key, mod_date=None, sum_closed='right', other_closed='left', discrete=True, station_attrs=None, station_encoding=None):
     """
 
     """
@@ -574,6 +744,17 @@ def prepare_results(data_dict, dataset_list, station_dict, data_df, run_date_key
 
     tz_str = 'Etc/GMT{0:+}'
 
+    ## Determine index
+    if not isinstance(results_data.index, pd.MultiIndex):
+        raise TypeError('The data_df index must be a MultiIndex.')
+
+    data_index = list(results_data.index.names)
+
+    if not 'time' in data_index:
+        raise ValueError('time must be in the data_df index.')
+
+    other_index = [i for i in data_index if i != 'time']
+
     ## Iterate through each dataset
     for ds in dataset_list:
         # print(ds['dataset_id'])
@@ -597,7 +778,10 @@ def prepare_results(data_dict, dataset_list, station_dict, data_df, run_date_key
         utc_offset = ds_mapping['utc_offset']
         parameter = ds_mapping['parameter']
 
-        ts_data1 = data_df.copy()
+        main_cols = data_index.copy()
+        main_cols.extend([parameter])
+
+        ts_data1 = results_data.reset_index().copy()
 
         # Convert times to local TZ if necessary
         if (not freq_code in ['T', 'H', '1H']) and (not utc_offset == '0H'):
@@ -608,16 +792,16 @@ def prepare_results(data_dict, dataset_list, station_dict, data_df, run_date_key
         ## Aggregate data if necessary
         # Parameter
         if freq_code == 'T':
-            grp1 = ts_data1.groupby(['time', 'height'])
+            grp1 = ts_data1.groupby(data_index)
             data1 = grp1[parameter].mean()
 
         else:
             agg_fun = agg_stat_mapping[ds_mapping['aggregation_statistic']]
 
             if agg_fun == 'sum':
-                data1 = grp_ts_agg(ts_data1[['time', 'height', parameter]], 'height', 'time', freq_code, agg_fun, closed=sum_closed)
+                data1 = grp_ts_agg(ts_data1[main_cols], other_index, 'time', freq_code, agg_fun, closed=sum_closed)
             else:
-                data1 = grp_ts_agg(ts_data1[['time', 'height', parameter]], 'height', 'time', freq_code, agg_fun, discrete, closed=other_closed)
+                data1 = grp_ts_agg(ts_data1[main_cols], other_index, 'time', freq_code, agg_fun, discrete, closed=other_closed)
 
         # Quality code
         if qual_col in ts_data1.columns:
@@ -625,7 +809,9 @@ def prepare_results(data_dict, dataset_list, station_dict, data_df, run_date_key
             if freq_code == 'T':
                 qual1 = grp1[qual_col].min()
             else:
-                qual1 = grp_ts_agg(ts_data1[['time', 'height', qual_col]], 'height', 'time', freq_code, 'min')
+                qual_cols = data_index.copy()
+                qual_cols.extend([qual_col])
+                qual1 = grp_ts_agg(ts_data1[qual_cols], other_index, 'time', freq_code, 'min')
             df3 = pd.concat([data1, qual1], axis=1).reset_index().dropna()
 
             ancillary_variables.extend([qual_col])
@@ -642,9 +828,9 @@ def prepare_results(data_dict, dataset_list, station_dict, data_df, run_date_key
         if (not freq_code in ['T', 'H', '1H']) and (not utc_offset == '0H'):
             df4['time'] = df4['time'].dt.tz_localize(tz1).dt.tz_convert('utc').dt.tz_localize(None)
 
-        df4.set_index(['time', 'height'], inplace=True)
+        df4.set_index(data_index, inplace=True)
 
-        new1 = data_to_xarray(df4, station_dict, parameter, attrs1, encoding1, station_attrs=station_attrs, station_encoding=station_encoding, run_date=run_date_key, ancillary_variables=ancillary_variables, compression='zstd')
+        new1 = data_to_xarray(df4, station_data, parameter, attrs1, encoding1, station_attrs=station_attrs, station_encoding=station_encoding, run_date=run_date_key, ancillary_variables=ancillary_variables, compression='zstd')
 
         ## Update the data_dict
         ds_id = ds_mapping['dataset_id']
@@ -652,167 +838,6 @@ def prepare_results(data_dict, dataset_list, station_dict, data_df, run_date_key
         data_dict[ds_id].append(new1)
 
 
-def split_grid(arr, x_size, y_size, x_name='x', y_name='y'):
-    """
-    Function to split an n-dimensional dataset along the x and y dimensions.
-
-    Parameters
-    ----------
-    arr : DataArray
-        An xarray DataArray with at least x and y dimensions. It can have any number of dimensions, though it probably does not make much sense to have greater than 4 dimensions.
-    x_size : int
-        The size or length of the smaller grids in the x dimension.
-    y_size : int
-        The size or length of the smaller grids in the y dimension.
-    x_name : str
-        The x dimension name.
-    y_name : str
-        The y dimension name.
-
-    Returns
-    -------
-    List of DataArrays
-        The result contains none of the original attributes.
-    """
-    ## Get the dimension data
-    dims = arr.dims
-    x_index = dims.index(x_name)
-    y_index = dims.index(y_name)
-    data_name = arr.name
-
-    arr_shape = arr.shape
-
-    m = arr_shape[x_index]
-    n = arr_shape[y_index]
-    dtype = arr.dtype
-
-    ## Build the new regular array to be queried
-    y_diff = arr[y_name].diff(y_name, 1).median().values
-    x_diff = arr[x_name].diff(x_name, 1).median().values
-
-    bpx = ((m-1)//x_size + 1) # blocks per x
-    bpy = ((n-1)//y_size + 1) # blocks per y
-    M = x_size * bpx
-    N = y_size * bpy
-
-    x_y = list(arr_shape)
-    x_y[x_index] = M
-    x_y[y_index] = N
-
-    sel1 = tuple(slice(0, s) for s in arr_shape)
-
-    A = np.nan * np.ones(x_y)
-    A[sel1] = arr
-
-    # x array
-    x_start = arr[x_name][0].values
-    x_int = M * x_diff
-    x_end = x_start + x_int
-    xs = np.arange(x_start, x_end, x_diff)
-
-    # y array
-    y_start = arr[y_name][0].values
-    y_int = M * y_diff
-    y_end = y_start + y_int
-    ys = np.arange(y_start, y_end, y_diff)
-
-    # Coords
-    coords = []
-    new_dims = []
-    for d in dims:
-        name = d
-        if d == x_name:
-            c = xs
-        elif d == y_name:
-            c = ys
-        else:
-            c = arr[d]
-        coords.extend([c])
-        new_dims.extend([name])
-
-    # New DataArray
-    A1 = xr.DataArray(A, coords=coords, dims=new_dims, name=data_name)
-
-    block_list = []
-    previous_x = 0
-    for x_block in range(bpy):
-        previous_x = x_block * x_size
-        previous_y = 0
-        for y_block in range(bpx):
-            previous_y = y_block * y_size
-            x_slice = slice(previous_x, previous_x+x_size)
-            y_slice = slice(previous_y, previous_y+y_size)
-
-            sel2 = list(sel1)
-            sel2[x_index] = x_slice
-            sel2[y_index] = y_slice
-
-            block = A1[tuple(sel2)]
-
-            # remove nans
-            block = block.dropna(y_name, 'all')
-            block = block.dropna(x_name, 'all')
-
-            ## append
-            if block.size:
-                block_list.append(block.astype(dtype))
-
-    return block_list
-
-
-def determine_array_size(arr, starting_x_size=100, starting_y_size=100, increment=100, min_size=800, max_size=1100, x_name='x', y_name='y'):
-    """
-    Function to determine the appropriate grid size for splitting.
-
-    Parameters
-    ----------
-    arr : DataArray
-        An xarray DataArray with at least x and y dimensions. It can have any number of dimensions, though it probably does not make much sense to have greater than 4 dimensions.
-    starting_x_size : int
-        The initial size or length of the smaller grids in the x dimension.
-    starting_y_size : int
-        The initial size or length of the smaller grids in the y dimension.
-    increment : int
-        The incremental grid size to be added iteratively to the starting sizes.
-    min_size : int
-        The minimum acceptable object size in KB.
-    max_size : int
-        The maximum acceptable object size in KB.
-    x_name : str
-        The x dimension name.
-    y_name : str
-        The y dimension name.
-
-    Returns
-    -------
-    dict
-        Of the optimised grid size results.
-    """
-    max_obj_size = 0
-    x_size = starting_x_size
-    y_size = starting_y_size
-
-    while True:
-        block_list = split_grid(arr, x_size=x_size, y_size=y_size, x_name=x_name, y_name=y_name)
-        obj_sizes = [len(write_pkl_zstd(nc.to_netcdf())) for nc in block_list]
-        max_obj_size = max(obj_sizes)
-
-        if max_obj_size < min_size*1000:
-            x_size = x_size + increment
-            y_size = y_size + increment
-        else:
-            break
-
-    if max_obj_size > max_size*1000:
-        print('max_object_size:', str(max_obj_size))
-        raise ValueError('max object size is greater than the allotted size. Reduce the increment value and start again.')
-
-    obj_dict = {'x_size': x_size, 'y_size': y_size, 'max_obj_size': max_obj_size, 'min_obj_size': min(obj_sizes)}
-
-    return obj_dict
-
-
-
 
 
 
